@@ -1,86 +1,83 @@
-from __future__ import unicode_literals
-
 import warnings
 from collections import defaultdict
 from copy import deepcopy
-from functools import partial
+from inspect import getfullargspec
+from sys import exc_info
 from threading import local
+from typing import Callable, List, Optional, Dict, Union, Sequence, Any, Tuple
 
-from django import VERSION
 from django.conf import settings
 from django.core.cache import cache
-from django.db.models import signals
+from django.db.models import signals, QuerySet
 from django.template.base import (
-    FilterExpression, Lexer, Parser, Token, Variable, VariableDoesNotExist, TOKEN_BLOCK, UNKNOWN_SOURCE, TOKEN_TEXT,
-    TOKEN_VAR, VARIABLE_TAG_START)
+    FilterExpression, Lexer, Parser, Token, Variable, VariableDoesNotExist, VARIABLE_TAG_START, Context)
 from django.template.defaulttags import url as url_tag
 from django.template.loader import get_template
-from django.utils import six
-from django.utils.encoding import python_2_unicode_compatible
+from django.utils import module_loading
 from django.utils.http import urlquote
 from django.utils.translation import get_language
 
+from .compat import TOKEN_BLOCK, TOKEN_TEXT, TOKEN_VAR
 from .exceptions import SiteTreeError
 from .settings import (
     ALIAS_TRUNK, ALIAS_THIS_CHILDREN, ALIAS_THIS_SIBLINGS, ALIAS_THIS_PARENT_SIBLINGS, ALIAS_THIS_ANCESTOR_CHILDREN,
-    UNRESOLVED_ITEM_MARKER, RAISE_ITEMS_ERRORS_ON_DEBUG, CACHE_TIMEOUT, DYNAMIC_ONLY)
+    UNRESOLVED_ITEM_MARKER, RAISE_ITEMS_ERRORS_ON_DEBUG, CACHE_TIMEOUT, DYNAMIC_ONLY, ADMIN_APP_NAME, SITETREE_CLS)
 from .utils import get_tree_model, get_tree_item_model, import_app_sitetree_module, generate_id_for
 
 if False:  # pragma: nocover
-    from django.template import Context
-    from .models import TreeItemBase
-
-
-if VERSION >= (1, 9, 0):
-    get_lexer = partial(Lexer)
-else:
-    get_lexer = partial(Lexer, origin=UNKNOWN_SOURCE)
+    from django.contrib.auth.models import User  # noqa
+    from .models import TreeItemBase, TreeBase
 
+TypeDynamicTrees = Dict[str, Union[Dict[str, List['TreeBase']], List['TreeBase']]]
 
 MODEL_TREE_CLASS = get_tree_model()
 MODEL_TREE_ITEM_CLASS = get_tree_item_model()
 
 
-_ITEMS_PROCESSOR = None
+_ITEMS_PROCESSOR: Optional[Callable] = None
 """Stores tree items processor callable or None."""
 
-_I18N_TREES = []
+_ITEMS_PROCESSOR_ARGS_LEN: int = 0
+"""Number of arguments accepted by items processor."""
+
+_I18N_TREES: List[str] = []
 """Stores aliases of trees supporting internationalization."""
 
-_DYNAMIC_TREES = {}
+_DYNAMIC_TREES: TypeDynamicTrees = {}
 """Holds trees dynamically loaded from project apps."""
 
-_IDX_ORPHAN_TREES = 'orphans'
+_IDX_ORPHAN_TREES: str = 'orphans'
 """Dictionary index in `_DYNAMIC_TREES` for orphaned trees list."""
 
-_IDX_TPL = '%s|:|%s'
+_IDX_TPL: str = '%s|:|%s'
 """Name template used as dictionary index in `_DYNAMIC_TREES`."""
 
 _THREAD_LOCAL = local()
-_THREAD_SITETREE = 'sitetree'
-
-_CONTEXT_FLATTEN = VERSION >= (1, 11)
-
+_THREAD_SITETREE: str = 'sitetree'
 _UNSET = set()  # Sentinel
 
 
-def get_sitetree():
+def get_sitetree() -> 'SiteTree':
     """Returns SiteTree (thread-singleton) object, implementing utility methods.
+    This can return the built-in or a customized (see SITETREE_CLS setting) sitetree handler.
 
-    :rtype: SiteTree
     """
     sitetree = getattr(_THREAD_LOCAL, _THREAD_SITETREE, None)
 
     if sitetree is None:
-        sitetree = SiteTree()
+        sitetree = _SITETREE_CLS()
         setattr(_THREAD_LOCAL, _THREAD_SITETREE, sitetree)
 
     return sitetree
 
 
-def register_items_hook(func):
+def register_items_hook(func: Callable):
     """Registers a hook callable to process tree items right before they are passed to templates.
 
+    .. deprecated:: 1.13.0
+       Items hooking with `register_items_hook` is deprecated, please use `SITETREE_CLS`
+       setting customization instead and override .apply_hook() method.
+
     Callable should be able to:
 
         a) handle ``tree_items`` and ``tree_sender`` key params.
@@ -104,7 +101,7 @@ def register_items_hook(func):
             if tree_sender == 'menu.children':
                 # Lets add 'Hooked: ' to resolved titles of every item.
                 for item in tree_items:
-                    item.title_resolved = 'Hooked: %s' % item.title_resolved
+                    item.title_resolved = f'Hooked: {item.title_resolved}'
             # Return items list mutated or not.
             return tree_items
 
@@ -113,11 +110,24 @@ def register_items_hook(func):
 
     :param func:
     """
+    warnings.warn(
+        'Items hooking with `register_items_hook` is deprecated, '
+        'please use `SITETREE_CLS` settings customization instead.',
+        DeprecationWarning, 2)
+
     global _ITEMS_PROCESSOR
+    global _ITEMS_PROCESSOR_ARGS_LEN
+
     _ITEMS_PROCESSOR = func
 
+    if func:
+        args_len = len(getfullargspec(func).args)
+        if args_len not in {2, 3}:
+            raise SiteTreeError('`register_items_hook()` expects a function with two or three arguments.')
+        _ITEMS_PROCESSOR_ARGS_LEN = args_len
 
-def register_i18n_trees(aliases):
+
+def register_i18n_trees(aliases: List[str]):
     """Registers aliases of internationalized sitetrees.
     Internationalized sitetrees are those, which are dubbed by other trees having
     locale identifying suffixes in their aliases.
@@ -141,12 +151,17 @@ def register_i18n_trees(aliases):
         register_i18n_trees(['my_tree', 'my_another_tree'])
 
     :param aliases:
+
     """
     global _I18N_TREES
     _I18N_TREES = aliases
 
 
-def register_dynamic_trees(trees, *args, **kwargs):
+def register_dynamic_trees(
+        trees: Union[List['TreeBase'], Dict[str, List['TreeBase']]],
+        *args,
+        **kwargs
+):
     """Registers dynamic trees to be available for `sitetree` runtime.
     Expects `trees` to be an iterable with structures created with `compose_dynamic_tree()`.
 
@@ -187,7 +202,7 @@ def register_dynamic_trees(trees, *args, **kwargs):
         trees = [trees]
         trees.extend(args)
 
-    for tree in trees:
+    for tree in trees or []:
         if tree is not None and tree['sitetrees'] is not None:
             if tree['tree'] is None:
                 # Register trees as they are defined in app.
@@ -209,28 +224,32 @@ def register_dynamic_trees(trees, *args, **kwargs):
         cache_.reset()
 
 
-def get_dynamic_trees():
+def get_dynamic_trees() -> TypeDynamicTrees:
     """Returns a dictionary with currently registered dynamic trees."""
     return _DYNAMIC_TREES
 
 
-def compose_dynamic_tree(src, target_tree_alias=None, parent_tree_item_alias=None, include_trees=None):
+def compose_dynamic_tree(
+        src: Union[str, Sequence['TreeBase'], Sequence['TreeItemBase']],
+        target_tree_alias: str = None,
+        parent_tree_item_alias: str = None,
+        include_trees: List[str] = None
+) -> dict:
     """Returns a structure describing a dynamic sitetree.utils
     The structure can be built from various sources,
 
-    :param str|iterable src: If a string is passed to `src`, it'll be treated as the name of an app,
+    :param src: If a string is passed to `src`, it'll be treated as the name of an app,
         from where one want to import sitetrees definitions. `src` can be an iterable
         of tree definitions (see `sitetree.toolbox.tree()` and `item()` functions).
 
-    :param str|unicode target_tree_alias: Static tree alias to attach items from dynamic trees to.
+    :param target_tree_alias: Static tree alias to attach items from dynamic trees to.
 
-    :param str|unicode parent_tree_item_alias: Tree item alias from a static tree to attach items from dynamic trees to.
+    :param parent_tree_item_alias: Tree item alias from a static tree to attach items from dynamic trees to.
 
-    :param list include_trees: Sitetree aliases to filter `src`.
+    :param include_trees: Sitetree aliases to filter `src`.
 
-    :rtype: dict
     """
-    def result(sitetrees=src):
+    def result(sitetrees):
         if include_trees is not None:
             sitetrees = [tree for tree in sitetrees if tree.alias in include_trees]
 
@@ -240,7 +259,7 @@ def compose_dynamic_tree(src, target_tree_alias=None, parent_tree_item_alias=Non
             'tree': target_tree_alias,
             'parent_item': parent_tree_item_alias}
 
-    if isinstance(src, six.string_types):
+    if isinstance(src, str):
         # Considered to be an application name.
         try:
             module = import_app_sitetree_module(src)
@@ -248,25 +267,24 @@ def compose_dynamic_tree(src, target_tree_alias=None, parent_tree_item_alias=Non
 
         except ImportError as e:
             if settings.DEBUG:
-                warnings.warn('Unable to register dynamic sitetree(s) for `%s` application: %s. ' % (src, e))
-            return None
+                warnings.warn(f'Unable to register dynamic sitetree(s) for `{src}` application: {e}. ')
+            return {}
 
-    return result()
+    return result(src)
 
 
-@python_2_unicode_compatible
-class LazyTitle(object):
+class LazyTitle:
     """Lazily resolves any variable found in a title of an item.
-    Produces resolved title as unicode representation."""
+    Produces resolved title as unicode representation.
 
-    def __init__(self, title):
-        """
-        :param str|unicode title:
-        """
+    """
+    __slots__ = ['title']
+
+    def __init__(self, title: str):
         self.title = title
 
     def __str__(self):
-        my_lexer = get_lexer(self.title)
+        my_lexer = Lexer(self.title)
         my_tokens = my_lexer.tokenize()
 
         # Deliberately strip off template tokens that are not text or variable.
@@ -281,11 +299,11 @@ class LazyTitle(object):
         return self.__str__() == other
 
 
-class Cache(object):
+class Cache:
     """Contains cache-related stuff."""
 
     def __init__(self):
-        self.cache = None
+        self.cache: dict = {}
 
         cache_empty = self.empty
         # Listen for signals from the models.
@@ -294,6 +312,7 @@ class Cache(object):
         signals.post_delete.connect(cache_empty, sender=MODEL_TREE_ITEM_CLASS)
         # Listen to the changes in item permissions table.
         signals.m2m_changed.connect(cache_empty, sender=MODEL_TREE_ITEM_CLASS.access_permissions)
+
         self.init()
 
     @classmethod
@@ -325,49 +344,54 @@ class Cache(object):
 
         kwargs.get('init', True) and self.init()
 
-    def get_entry(self, entry_name, key):
+    def get_entry(self, entry_name: str, key) -> Any:
         """Returns cache entry parameter value by its name.
 
-        :param str|unicode entry_name:
+        :param entry_name:
         :param key:
-        :return:
+
         """
         return self.cache[entry_name].get(key, False)
 
-    def update_entry_value(self, entry_name, key, value):
+    def update_entry_value(self, entry_name: str, key: str, value: Any):
         """Updates cache entry parameter with new data.
 
-        :param str|unicode entry_name:
+        :param entry_name:
         :param key:
         :param value:
+
         """
         if key not in self.cache[entry_name]:
             self.cache[entry_name][key] = {}
 
         self.cache[entry_name][key].update(value)
 
-    def set_entry(self, entry_name, key, value):
+    def set_entry(self, entry_name: str, key: str, value: Any):
         """Replaces entire cache entry parameter data by its name with new data.
 
-        :param str|unicode entry_name:
+        :param entry_name:
         :param key:
         :param value:
+
         """
         self.cache[entry_name][key] = value
 
 
-class SiteTree(object):
+class SiteTree:
     """Main logic handler."""
 
+    cache_cls = Cache  # Allow customizations.
+
     def __init__(self):
         self.init(context=None)
 
-    def init(self, context):
+    def init(self, context: Optional[Context]):
         """Initializes sitetree to handle new request.
 
-        :param Context|None context:
+        :param context:
+
         """
-        self.cache = Cache()
+        self.cache = self.cache_cls()
         self.current_page_context = context
         self.current_request = context.get('request', None) if context else None
         self.current_lang = get_language()
@@ -377,19 +401,19 @@ class SiteTree(object):
         self._items_urls = {}  # Resolved urls are cache for a request.
         self._current_items = {}
 
-    def resolve_tree_i18n_alias(self, alias):
+    def resolve_tree_i18n_alias(self, alias: str) -> str:
         """Resolves internationalized tree alias.
         Verifies whether a separate sitetree is available for currently active language.
         If so, returns i18n alias. If not, returns the initial alias.
 
-        :param str|unicode alias:
-        :rtype: str|unicode
+        :param alias:
+
         """
         if alias not in _I18N_TREES:
             return alias
 
         current_language_code = self.current_lang
-        i18n_tree_alias = '%s_%s' % (alias, current_language_code)
+        i18n_tree_alias = f'{alias}_{current_language_code}'
         trees_count = self.cache.get_entry('tree_aliases', i18n_tree_alias)
 
         if trees_count is False:
@@ -402,16 +426,19 @@ class SiteTree(object):
         return alias
 
     @staticmethod
-    def attach_dynamic_tree_items(tree_alias, src_tree_items):
+    def attach_dynamic_tree_items(
+            tree_alias: str,
+            src_tree_items: Union[Sequence['TreeItemBase'], QuerySet]
+    ) -> List['TreeItemBase']:
         """Attaches dynamic sitetrees items registered with `register_dynamic_trees()`
         to an initial (source) items list.
 
-        :param str|unicode tree_alias:
-        :param list src_tree_items:
-        :rtype: list
+        :param tree_alias:
+        :param src_tree_items:
+
         """
         if not _DYNAMIC_TREES:
-            return src_tree_items
+            return list(src_tree_items)
 
         # This guarantees that a dynamic source stays intact,
         # no matter how dynamic sitetrees are attached.
@@ -438,7 +465,8 @@ class SiteTree(object):
 
                 for tree in trees[idx]:
                     tree.alias = tree_alias
-                    for dyn_item in tree.dynamic_items:
+
+                    for dyn_item in tree.dynamic_items:  # noqa dynamic attr
                         if dyn_item.parent is None:
                             dyn_item.parent = static_item
                         # Unique IDs are required for the same trees attached
@@ -452,16 +480,15 @@ class SiteTree(object):
                 trees = deepcopy(_DYNAMIC_TREES)
                 for tree in trees[idx]:
                     tree.alias = tree_alias
-                    items.extend(tree.dynamic_items)
+                    items.extend(tree.dynamic_items)  # noqa dynamic attr
 
         return items
 
-    def current_app_is_admin(self):
-        """Returns boolean whether current application is Admin contrib.
+    def current_app_is_admin(self) -> bool:
+        """Returns boolean whether current application is Admin contrib."""
 
-        :rtype: bool
-        """
         is_admin = self._current_app_is_admin
+
         if is_admin is None:
             context = self.current_page_context
 
@@ -474,18 +501,18 @@ class SiteTree(object):
             if current_app is None:  # Try from global context dict.
                 current_app = context.get('current_app', '')
 
-            is_admin = current_app == 'admin'
+            is_admin = current_app == ADMIN_APP_NAME
             self._current_app_is_admin = is_admin
 
         return is_admin
 
-    def get_sitetree(self, alias):
+    def get_sitetree(self, alias: str) -> Tuple[str, List['TreeItemBase']]:
         """Gets site tree items from the given site tree.
         Caches result to dictionary.
         Returns (tree alias, tree items) tuple.
 
-        :param str|unicode alias:
-        :rtype: tuple
+        :param alias:
+
         """
         cache_ = self.cache
         get_cache_entry = cache_.get_entry
@@ -526,15 +553,19 @@ class SiteTree(object):
         # Prepare items by ids cache if needed.
         if caching_required:
             # We need this extra pass to avoid future problems on items depth calculation.
+            cache_update = cache_.update_entry_value
             for item in sitetree:
-                cache_.update_entry_value('items_by_ids', alias, {item.id: item})
+                cache_update('items_by_ids', alias, {item.id: item})
+
+        url = self.url
+        calculate_item_depth = self.calculate_item_depth
 
         for item in sitetree:
             if caching_required:
                 item.has_children = False
 
                 if not hasattr(item, 'depth'):
-                    item.depth = self.calculate_item_depth(alias, item.id)
+                    item.depth = calculate_item_depth(alias, item.id)
                 item.depth_range = range(item.depth)
 
                 # Resolve item permissions.
@@ -544,10 +575,10 @@ class SiteTree(object):
                         else item.access_permissions.all())
 
                     item.perms = set(
-                        ['%s.%s' % (perm.content_type.app_label, perm.codename) for perm in permissions_src])
+                        [f'{perm.content_type.app_label}.{perm.codename}' for perm in permissions_src])
 
             # Contextual properties.
-            item.url_resolved = self.url(item)
+            item.url_resolved = url(item)
             item.title_resolved = LazyTitle(item.title) if VARIABLE_TAG_START in item.title else item.title
             item.is_current = False
             item.in_current_branch = False
@@ -561,39 +592,40 @@ class SiteTree(object):
 
         return alias, sitetree
 
-    def calculate_item_depth(self, tree_alias, item_id, depth=0):
+    def calculate_item_depth(self, tree_alias: str, item_id: int, depth: int = 0):
         """Calculates depth of the item in the tree.
 
-        :param str|unicode tree_alias:
-        :param int item_id:
-        :param int depth:
-        :rtype: int
+        :param tree_alias:
+        :param item_id:
+        :param depth:
+
         """
         item = self.get_item_by_id(tree_alias, item_id)
 
         if hasattr(item, 'depth'):
             depth = item.depth + depth
+
         else:
             if item.parent is not None:
                 depth = self.calculate_item_depth(tree_alias, item.parent.id, depth + 1)
 
         return depth
 
-    def get_item_by_id(self, tree_alias, item_id):
+    def get_item_by_id(self, tree_alias: str, item_id: int) -> 'TreeItemBase':
         """Get the item from the tree by its ID.
 
-        :param str|unicode tree_alias:
-        :param int item_id:
-        :rtype: TreeItemBase
+        :param tree_alias:
+        :param item_id:
+
         """
         return self.cache.get_entry('items_by_ids', tree_alias)[item_id]
 
-    def get_tree_current_item(self, tree_alias):
+    def get_tree_current_item(self, tree_alias: str) -> Optional['TreeItemBase']:
         """Resolves current tree item of 'tree_alias' tree matching current
         request path against URL of given tree item.
 
-        :param str|unicode tree_alias:
-        :rtype: TreeItemBase
+        :param tree_alias:
+
         """
         current_item = self._current_items.get(tree_alias, _UNSET)
 
@@ -602,7 +634,7 @@ class SiteTree(object):
             if current_item is not None:
                 current_item.is_current = True  # Could be reset by .get_sitetree()
 
-            return current_item
+            return current_item  # noqa
 
         current_item = None
 
@@ -631,20 +663,20 @@ class SiteTree(object):
 
         return current_item
 
-    def url(self, sitetree_item, context=None):
+    def url(self, sitetree_item: Union['TreeItemBase', FilterExpression], context: Context = None) -> str:
         """Resolves item's URL.
 
-        :param TreeItemBase sitetree_item: TreeItemBase heir object, 'url' property of which
+        :param sitetree_item: TreeItemBase heir object, 'url' property of which
             is processed as URL pattern or simple URL.
 
-        :param Context context:
+        :param context:
 
-        :rtype: str|unicode
         """
         context = context or self.current_page_context
+        resolve_var = self.resolve_var
 
         if not isinstance(sitetree_item, MODEL_TREE_ITEM_CLASS):
-            sitetree_item = self.resolve_var(sitetree_item, context)
+            sitetree_item = resolve_var(sitetree_item, context)
 
         resolved_url = self._items_urls.get(sitetree_item)
         if resolved_url is not None:
@@ -660,25 +692,20 @@ class SiteTree(object):
                 view_path = url.split(' ')
                 # We should try to resolve URL parameters from site tree item.
                 for view_argument in view_path[1:]:
-                    resolved = self.resolve_var(view_argument)
-                    # In case of non-ascii data we leave variable unresolved.
-                    if isinstance(resolved, six.text_type):
-                        if resolved.encode('ascii', 'ignore').decode('ascii') != resolved:
-                            resolved = view_argument
-
+                    resolved = resolve_var(view_argument)
                     # We enclose arg in double quotes as already resolved.
-                    all_arguments.append('"%s"' % resolved)
+                    all_arguments.append(f'"{resolved}"')
 
                 view_path = view_path[0].strip('"\' ')
 
-            url_pattern = "'%s' %s" % (view_path, ' '.join(all_arguments))
+            url_pattern = f"'{view_path}' {' '.join(all_arguments)}"
 
         else:
-            url_pattern = '%s' % sitetree_item.url
+            url_pattern = f'{sitetree_item.url}'
 
         if sitetree_item.urlaspattern:
             # Form token to pass to Django 'url' tag.
-            url_token = 'url %s as item.url_resolved' % url_pattern
+            url_token = f'url {url_pattern} as item.url_resolved'
             url_tag(
                 Parser(None),
                 Token(token_type=TOKEN_BLOCK, contents=url_token)
@@ -693,20 +720,31 @@ class SiteTree(object):
 
         return resolved_url
 
-    def init_tree(self, tree_alias, context):
+    def init_tree(
+            self,
+            tree_alias: str,
+            context: Context
+    ) -> Tuple[Optional[str], Optional[List['TreeItemBase']]]:
         """Initializes sitetree in memory.
 
         Returns tuple with resolved tree alias and items on success.
 
         On fail returns (None, None).
 
-        :param str|unicode tree_alias:
-        :param Context context:
-        :rtype: tuple
+        :param tree_alias:
+        :param context:
+
         """
         request = context.get('request', None)
 
         if request is None:
+
+            if any(exc_info()):
+                # Probably we're in a technical
+                # exception handling view. So we won't mask
+                # the initial exception with the one below.
+                return None, None
+
             raise SiteTreeError(
                 'Sitetree requires "django.core.context_processors.request" template context processor to be active. '
                 'If it is, check that your view pushes request data into the template.')
@@ -723,22 +761,22 @@ class SiteTree(object):
 
         return tree_alias, sitetree_items
 
-    def get_current_page_title(self, tree_alias, context):
+    def get_current_page_title(self, tree_alias: str, context: Context) -> str:
         """Returns resolved from sitetree title for current page.
 
-        :param str|unicode tree_alias:
-        :param Context context:
-        :rtype: str|unicode
+        :param tree_alias:
+        :param context:
+
         """
         return self.get_current_page_attr('title_resolved', tree_alias, context)
 
-    def get_current_page_attr(self, attr_name, tree_alias, context):
+    def get_current_page_attr(self, attr_name: str, tree_alias: str, context: Context) -> str:
         """Returns an arbitrary attribute of a sitetree item resolved as current for current page.
 
-        :param str|unicode attr_name:
-        :param str|unicode tree_alias:
-        :param Context context:
-        :rtype: str|unicode
+        :param attr_name:
+        :param tree_alias:
+        :param context:
+
         """
         tree_alias, sitetree_items = self.init_tree(tree_alias, context)
         current_item = self.get_tree_current_item(tree_alias)
@@ -746,19 +784,19 @@ class SiteTree(object):
         if current_item is None:
             if settings.DEBUG and RAISE_ITEMS_ERRORS_ON_DEBUG:
                 raise SiteTreeError(
-                    'Unable to resolve current sitetree item to get a `%s` for current page. Check whether '
-                    'there is an appropriate sitetree item defined for current URL.' % attr_name)
+                    f'Unable to resolve current sitetree item to get a `{attr_name}` for current page. Check whether '
+                    'there is an appropriate sitetree item defined for current URL.')
 
             return ''
 
         return getattr(current_item, attr_name, '')
 
-    def get_ancestor_level(self, current_item, depth=1):
+    def get_ancestor_level(self, current_item: 'TreeItemBase', depth: int = 1) -> 'TreeItemBase':
         """Returns ancestor of level `deep` recursively
 
-        :param TreeItemBase current_item:
-        :param int depth:
-        :rtype: TreeItemBase
+        :param current_item:
+        :param depth:
+
         """
         if current_item.parent is None:
             return current_item
@@ -768,18 +806,18 @@ class SiteTree(object):
 
         return self.get_ancestor_level(current_item.parent, depth=depth-1)
 
-    def menu(self, tree_alias, tree_branches, context):
+    def menu(self, tree_alias: str, tree_branches: str, context: Context) -> List['TreeItemBase']:
         """Builds and returns menu structure for 'sitetree_menu' tag.
 
-        :param str|unicode tree_alias:
-        :param str|unicode tree_branches:
-        :param Context context:
-        :rtype: list|str
+        :param tree_alias:
+        :param tree_branches:
+        :param context:
+
         """
         tree_alias, sitetree_items = self.init_tree(tree_alias, context)
 
         if not sitetree_items:
-            return ''
+            return []
 
         tree_branches = self.resolve_var(tree_branches)
 
@@ -819,9 +857,11 @@ class SiteTree(object):
             else:
                 parent_aliases.append(branch_id)
 
+        check_access = self.check_access
+
         menu_items = []
         for item in sitetree_items:
-            if not item.hidden and item.inmenu and self.check_access(item, context):
+            if not item.hidden and item.inmenu and check_access(item, context):
                 if item.parent is None:
                     if parent_isnull:
                         menu_items.append(item)
@@ -831,30 +871,37 @@ class SiteTree(object):
 
         menu_items = self.apply_hook(menu_items, 'menu')
         self.update_has_children(tree_alias, menu_items, 'menu')
+
         return menu_items
 
-    @classmethod
-    def apply_hook(cls, items, sender):
-        """Applies item processing hook, registered with ``register_item_hook()``
-        to items supplied, and returns processed list.
+    def apply_hook(self, items: List['TreeItemBase'], sender: str) -> List['TreeItemBase']:
+        """Applies a custom items processing hook to items supplied, and returns processed list.
+        Suitable for items filtering and other manipulations.
 
         Returns initial items list if no hook is registered.
 
-        :param list items:
-        :param str|unicode sender: menu, breadcrumbs, sitetree, {type}.children, {type}.has_children
-        :rtype: list
+        .. note:: Use `SITETREE_CLS` setting customization and override this method.
+
+        :param items:
+        :param sender: menu, breadcrumbs, sitetree, {type}.children, {type}.has_children
+
         """
-        if _ITEMS_PROCESSOR is None:
+        processor = _ITEMS_PROCESSOR
+
+        if processor is None:
             return items
 
-        return _ITEMS_PROCESSOR(tree_items=items, tree_sender=sender)
+        if _ITEMS_PROCESSOR_ARGS_LEN == 2:
+            return processor(tree_items=items, tree_sender=sender)
 
-    def check_access(self, item, context):
+        return processor(tree_items=items, tree_sender=sender, context=self.current_page_context)
+
+    def check_access(self, item: 'TreeItemBase', context: Context) -> bool:
         """Checks whether a current user has an access to a certain item.
 
-        :param TreeItemBase item:
-        :param Context context:
-        :rtype: bool
+        :param item:
+        :param context:
+
         """
         if hasattr(self.current_request.user.is_authenticated, '__call__'):
             authenticated = self.current_request.user.is_authenticated()
@@ -871,29 +918,38 @@ class SiteTree(object):
             user_perms = self._current_user_permissions
 
             if user_perms is _UNSET:
-                user_perms = set(context['user'].get_all_permissions())
+                user_perms = self.get_permissions(context['user'], item)
                 self._current_user_permissions = user_perms
 
             if item.access_perm_type == MODEL_TREE_ITEM_CLASS.PERM_TYPE_ALL:
-                if len(item.perms) != len(item.perms.intersection(user_perms)):
+                if len(item.perms) != len(item.perms.intersection(user_perms)):  # noqa dynamic attr
                     return False
             else:
-                if not len(item.perms.intersection(user_perms)):
+                if not len(item.perms.intersection(user_perms)):  # noqa dynamic attr
                     return False
 
         return True
 
-    def breadcrumbs(self, tree_alias, context):
+    def get_permissions(self, user: 'User', item: 'TreeItemBase') -> set:
+        """Returns a set of user and group level permissions for a given user.
+
+        :param user:
+        :param item:
+
+        """
+        return user.get_all_permissions()
+
+    def breadcrumbs(self, tree_alias: str, context: Context) -> List['TreeItemBase']:
         """Builds and returns breadcrumb trail structure for 'sitetree_breadcrumbs' tag.
 
-        :param str|unicode tree_alias:
-        :param Context context:
-        :rtype: list|str
+        :param tree_alias:
+        :param context:
+
         """
         tree_alias, sitetree_items = self.init_tree(tree_alias, context)
 
         if not sitetree_items:
-            return ''
+            return []
 
         current_item = self.get_tree_current_item(tree_alias)
 
@@ -924,17 +980,17 @@ class SiteTree(object):
 
         return items
 
-    def tree(self, tree_alias, context):
+    def tree(self, tree_alias: str, context:  Context) -> List['TreeItemBase']:
         """Builds and returns tree structure for 'sitetree_tree' tag.
 
-        :param str|unicode tree_alias:
-        :param Context context:
-        :rtype: list|str
+        :param tree_alias:
+        :param context:
+
         """
         tree_alias, sitetree_items = self.init_tree(tree_alias, context)
 
         if not sitetree_items:
-            return ''
+            return []
 
         tree_items = self.filter_items(self.get_children(tree_alias, None), 'sitetree')
         tree_items = self.apply_hook(tree_items, 'sitetree')
@@ -942,14 +998,20 @@ class SiteTree(object):
 
         return tree_items
 
-    def children(self, parent_item, navigation_type, use_template, context):
+    def children(
+            self,
+            parent_item: 'TreeItemBase',
+            navigation_type: str,
+            use_template: str,
+            context: Context
+    ) -> str:
         """Builds and returns site tree item children structure for 'sitetree_children' tag.
 
-        :param TreeItemBase parent_item:
-        :param str|unicode navigation_type: menu, sitetree
-        :param str|unicode use_template:
-        :param Context context:
-        :rtype: list
+        :param parent_item:
+        :param navigation_type: menu, sitetree
+        :param use_template:
+        :param context:
+
         """
         # Resolve parent item and current tree alias.
         parent_item = self.resolve_var(parent_item, context)
@@ -960,24 +1022,24 @@ class SiteTree(object):
 
         tree_items = self.get_children(tree_alias, parent_item)
         tree_items = self.filter_items(tree_items, navigation_type)
-        tree_items = self.apply_hook(tree_items, '%s.children' % navigation_type)
+        tree_items = self.apply_hook(tree_items, f'{navigation_type}.children')
         self.update_has_children(tree_alias, tree_items, navigation_type)
 
         my_template = get_template(use_template)
 
         context.push()
         context['sitetree_items'] = tree_items
-        rendered = my_template.render(context.flatten() if _CONTEXT_FLATTEN else context)
+        rendered = my_template.render(context.flatten())
         context.pop()
 
         return rendered
 
-    def get_children(self, tree_alias, item):
+    def get_children(self, tree_alias: str, item: Optional['TreeItemBase']) -> List['TreeItemBase']:
         """Returns item's children.
 
-        :param str|unicode tree_alias:
-        :param TreeItemBase|None item:
-        :rtype: list
+        :param tree_alias:
+        :param item:
+
         """
         if not self.current_app_is_admin():
             # We do not need i18n for a tree rendered in Admin dropdown.
@@ -985,27 +1047,32 @@ class SiteTree(object):
 
         return self.cache.get_entry('parents', tree_alias)[item]
 
-    def update_has_children(self, tree_alias, tree_items, navigation_type):
+    def update_has_children(self, tree_alias: str, tree_items: List['TreeItemBase'], navigation_type: str):
         """Updates 'has_children' attribute for tree items inplace.
 
-        :param str|unicode tree_alias:
-        :param list tree_items:
-        :param str|unicode navigation_type: sitetree, breadcrumbs, menu
+        :param tree_alias:
+        :param tree_items:
+        :param navigation_type: sitetree, breadcrumbs, menu
+
         """
+        get_children = self.get_children
+        filter_items = self.filter_items
+        apply_hook = self.apply_hook
+
         for tree_item in tree_items:
-            children = self.get_children(tree_alias, tree_item)
-            children = self.filter_items(children, navigation_type)
-            children = self.apply_hook(children, '%s.has_children' % navigation_type)
+            children = get_children(tree_alias, tree_item)
+            children = filter_items(children, navigation_type)
+            children = apply_hook(children, f'{navigation_type}.has_children')
             tree_item.has_children = len(children) > 0
 
-    def filter_items(self, items, navigation_type=None):
+    def filter_items(self, items: List['TreeItemBase'], navigation_type: str = None) -> List['TreeItemBase']:
         """Filters sitetree item's children if hidden and by navigation type.
 
         NB: We do not apply any filters to sitetree in admin app.
 
-        :param list items:
-        :param str|unicode navigation_type: sitetree, breadcrumbs, menu
-        :rtype: list
+        :param items:
+        :param navigation_type: sitetree, breadcrumbs, menu
+
         """
         if self.current_app_is_admin():
             return items
@@ -1022,19 +1089,19 @@ class SiteTree(object):
             if not check_access(item, context):
                 continue
 
-            if not getattr(item, 'in%s' % navigation_type, True):  # Hidden for current nav type
+            if not getattr(item, f'in{navigation_type}', True):  # Hidden for current nav type
                 continue
 
             items_filtered.append(item)
 
         return items_filtered
 
-    def get_ancestor_item(self, tree_alias, base_item):
+    def get_ancestor_item(self, tree_alias: str, base_item: 'TreeItemBase') -> 'TreeItemBase':
         """Climbs up the site tree to resolve root item for chosen one.
 
-        :param str|unicode tree_alias:
-        :param TreeItemBase base_item:
-        :rtype: TreeItemBase
+        :param tree_alias:
+        :param base_item:
+
         """
         parent = None
 
@@ -1046,30 +1113,36 @@ class SiteTree(object):
 
         return parent
 
-    def tree_climber(self, tree_alias, base_item):
+    def tree_climber(self, tree_alias: str, base_item: 'TreeItemBase'):
         """Climbs up the site tree to mark items of current branch.
 
-        :param str|unicode tree_alias:
-        :param TreeItemBase base_item:
+        :param tree_alias:
+        :param base_item:
+
         """
         if base_item is not None:
             base_item.in_current_branch = True
             if hasattr(base_item, 'parent') and base_item.parent is not None:
                 self.tree_climber(tree_alias, self.get_item_by_id(tree_alias, base_item.parent.id))
 
-    def resolve_var(self, varname, context=None):
+    def resolve_var(
+            self,
+            varname: Union[str, 'TreeItemBase', FilterExpression],
+            context: Context = None
+    ) -> Any:
         """Resolves name as a variable in a given context.
 
-        If no context specified page context' is considered as context.
+        If no context specified page context is considered as context.
+
+        :param varname:
+        :param context:
 
-        :param str|unicode varname:
-        :param Context context:
-        :return:
         """
         context = context or self.current_page_context
 
         if isinstance(varname, FilterExpression):
             varname = varname.resolve(context)
+
         else:
             varname = varname.strip()
 
@@ -1079,3 +1152,6 @@ class SiteTree(object):
                 varname = varname
 
         return varname
+
+
+_SITETREE_CLS = module_loading.import_string(SITETREE_CLS) if SITETREE_CLS else SiteTree
